#!/usr/bin/env python
# Copyright (c) 2021, Uwe Hernandez Acosta
#
# Distributed under the 3-clause BSD license, see accompanying file LICENSE
# or https://gitlab.hzdr.de/interpol/minterpy for details.

from setuptools import setup

setup()
